"""
.. module:: ttv2fast2furious
   :platform: Unix, Mac, Windows
   :synopsis: Routines for TTV analysis

.. moduleauthor:: Sam Hadden <samuel.hadden@cfa.harvard.edu>

"""
import warnings
import numpy as np
from scipy.optimize import lsq_linear
from scipy.special import gammainc,gammaincinv
from scipy.special import ellipk,ellipkinc,ellipe,ellipeinc
import scipy.integrate as integrate
from ttv2fast2furious.ttv_basis_functions import get_nearest_firstorder,get_superperiod,dt0_InnerPlanet,dt0_OuterPlanet,get_fCoeffs
from ttv2fast2furious.ttv_basis_functions import get_nearest_secondorder,dt2_InnerPlanet,dt2_OuterPlanet
from collections import OrderedDict

def ttv_basis_function_matrix_inner(P,P1,T0,T10,Ntrans,IncludeLinearBasis=True):
    """
    Compute the transit time basis function matrix, 'M', for a planet subject to an external perturber.

    Args:
        P (real): Inner planet period.
        P1 (real): Inner planet period.
        T (real): Outer planet initial time of transit
        T1 (real): Outer planet initial time of transit
        Ntrans (int):  number of transits 
        IncludeLinearBasis (bool): Whether basis functions representing a linear transit emphemris is included. Default is True.
           
    Returns:
        Array: the resulting basis function matrix is returned an (Ntrans, 5) array or an (Ntrans, 3) array if IncludeLinearBasis=False.
    """
    j = get_nearest_firstorder(P/P1)
    assert j > 0 , "Bad period ratio!!! P,P1 = %.3f \t %.3f"%(P,P1)
    superP = get_superperiod(P,P1)
    superPhase = 2*np.pi * (j * (-T10/P1) - (j-1) * (-T0/P) )
    Times = T0 + np.arange(Ntrans) * P

    # Normalize super-sin/cos so that basis fn amplitude is mu * (Zx/Zy)
    #
    j_2 = 1.0 / j / j
    Delta = (j-1) * P1 / P / j - 1
    Delta_2 = 1.0 / Delta / Delta
    alpha = (P/P1)**(2./3.)
    alpha_2 = 1.0 / alpha / alpha
    fCoeffIn,fCoeffOut=get_fCoeffs(j,alpha)
    denom = (fCoeffIn*fCoeffIn + fCoeffOut*fCoeffOut)**(0.5)
    S =  1.5 * (1-j) * alpha_2 * j_2 * Delta_2 * denom * P / (np.pi)
    C = -1.5 * (1-j) * alpha_2 * j_2 * Delta_2 * denom * P / (np.pi)

    dt0 = dt0_InnerPlanet(P,P1,T0,T10,Ntrans)
    superSin = S * np.sin(2*np.pi * Times / superP + superPhase)
    superCos = C * np.cos(2*np.pi * Times / superP + superPhase)
    
    if IncludeLinearBasis:
        basis_function_matrix=np.vstack((np.ones(Ntrans),np.arange(Ntrans),dt0,superSin,superCos)).T
    else:
        basis_function_matrix=np.vstack((dt0,superSin,superCos)).T

    return basis_function_matrix

def ttv_basis_function_matrix_outer(P,P1,T0,T10,Ntrans,IncludeLinearBasis=True):
    """
    Compute the transit time basis function matrix, 'M', for a planet subject to an interior perturber.

    Args:
        P (real): Inner planet period.
        P1 (real): Inner planet period.
        T (real): Outer planet initial time of transit
        T1 (real): Outer planet initial time of transit
        Ntrans (int):  number of transits 
        IncludeLinearBasis (bool): Whether basis functions representing a linear transit emphemris is included. Default is True.
           
    Returns:
        Array: the resulting basis function matrix is returned an (Ntrans, 5) array or an (Ntrans, 3) array if IncludeLinearBasis=False.
    """
    j = get_nearest_firstorder(P/P1)
    assert j > 0 , "Bad period ratio! P,P1 = %.3f \t %.3f"%(P,P1)
    superP = get_superperiod(P,P1)
    superPhase = 2*np.pi * (j * (-T10/P1) - (j-1) * (-T0/P) )
    Times = T10 + np.arange(Ntrans) * P1

    # Normalize super-sin/cos so that basis fn amplitude is mu * (Zx/Zy)
    #
    j_2 = 1.0 / j / j
    Delta = (j-1) * P1 / P / j - 1
    Delta_2 = 1.0 / Delta / Delta
    alpha = (P/P1)**(2./3.)
    alpha_2 = 1.0 / alpha / alpha
    fCoeffIn,fCoeffOut=get_fCoeffs(j,alpha)
    denom = (fCoeffIn*fCoeffIn + fCoeffOut*fCoeffOut)**(0.5)
    S1 =  1.5 * (j) * j_2 * Delta_2 * denom  * P1 / (np.pi)
    C1 = -1.5 * (j) * j_2 * Delta_2 * denom  * P1 / (np.pi)

    dt0 = dt0_OuterPlanet(P,P1,T0,T10,Ntrans)
    superSin = S1 * np.sin(2*np.pi * Times / superP + superPhase)
    superCos = C1 * np.cos(2*np.pi * Times / superP + superPhase)
    
    if IncludeLinearBasis:
        basis_function_matrix=np.vstack((np.ones(Ntrans),np.arange(Ntrans),dt0,superSin,superCos)).T
    else:
        basis_function_matrix=np.vstack((dt0,superSin,superCos)).T
    return basis_function_matrix
    
###################################
def PlanetPropertiestoLinearModelAmplitudes(T0,P,mass,e,varpi,T10,P1,mass1,e1,varpi1):
    """
    """

    j=get_nearest_firstorder(P/P1)
    j_2 = 1.0 / j / j
    Delta = (j-1) * P1 / P / j - 1
    Delta_2 = 1.0 / Delta / Delta

    ex = e * np.cos(varpi)
    ey = e * np.sin(varpi)
    ex1 = e1 * np.cos(varpi1)
    ey1 = e1 * np.sin(varpi1)

    alpha = (P/P1)**(2./3.)
    alpha_2 = 1.0 / alpha / alpha
    fCoeffIn,fCoeffOut=get_fCoeffs(j,alpha)
    denom = (fCoeffIn*fCoeffIn + fCoeffOut*fCoeffOut)**(0.5)
    Zx = fCoeffIn * ex + fCoeffOut * ex1
    Zy = fCoeffIn * ey + fCoeffOut * ey1
    Zx /= denom
    Zy /= denom

    S = mass1 * Zx
    C = mass1 * Zy
    S1 = mass * Zx
    C1 = mass * Zy
    return np.array([T0,P,mass1,S,C]),np.array([T10,P1,mass,S1,C1])
###################################
def get_ttv_basis_function_matrix(Pi,Pj,T0i,T0j,Ntrans):
    """
    Compute basis function matrix for planet `i`'s TTV due to planet `j`. 
    
    Arguments
    --------
    Pi: real
        TTV planet's period
    Pj: real
        Perturbing planet's period
    T0i: real
        TTV planet's intial time of transit
    """
    if Pi < Pj:
        return ttv_basis_function_matrix_inner(Pi,Pj,T0i,T0j,Ntrans,IncludeLinearBasis=False)
    else:
        return ttv_basis_function_matrix_outer(Pj,Pi,T0j,T0i,Ntrans,IncludeLinearBasis=False)
def get_linear_basis_basis_function_matrix(Ntransits):
    return np.vstack((np.ones(Ntransits),np.arange(Ntransits))).T
###################################
def MultiplanetSystemBasisFunctionMatrices(Nplanets,Periods,T0s,Ntransits,**kwargs):
    """
    Compute basis function matrices for the transit times of an `Nplanet` system.

    Parameters
    ----------
    Nplanets : int
        The number of transting planets to model.
    Periods  : ndarray
        Array listing planets' orbital periods.
    T0s      : ndarray
        Array listing the times of planets' first transits
    Ntransits: ndarray
        Array listing the number of transits to compute for each planet.

    Keyword Arguments
    -----------------
    ``InteractionMatrix``
        Specify the interactions between planets as a matrix. 
        By default, all planets are assumed to interact with one
        antoher.

    Returns
    -------
    list
        List of ndarrays with columns containing TTV basis functions of each planet.
    """
    InteractionMatrix=kwargs.get("InteractionMatrix",np.ones((Nplanets,Nplanets),dtype=bool))
    for i in range(Nplanets):
        # No self-interactions!
        InteractionMatrix[i,i]=False

    BasisFunctionMatrices=[get_linear_basis_basis_function_matrix(Nt) for Nt in Ntransits]
    for i in range(Nplanets):
        Pi = Periods[i]
        T0i = T0s[i]
        Ntransi = Ntransits[i]
        for j in range(Nplanets):
            if InteractionMatrix[i,j]:
                Pj = Periods[j]
                T0j = T0s[j]
                A = get_ttv_basis_function_matrix(Pi,Pj,T0i,T0j,Ntransi)
                BasisFunctionMatrices[i] = np.hstack((BasisFunctionMatrices[i],A))
            else:
                continue
    return BasisFunctionMatrices
###################################
def SetupInteractionMatrixWithMaxPeriodRatio(Periods, MaxRatio):
    Np=len(Periods)
    entries = []
    for P1 in Periods:
        for P2 in Periods:
            entries.append(np.max((P1/P2,P2/P1)) < MaxRatio)
    entries=np.array(entries).reshape((Np,Np))
    for i in range(Np):
        entries[i,i]=False
    return entries
###################################
def get_ttv_model_amplitudes(Pi,Pj,T0i,T0j,massi,massj,ei,ej,pmgi,pmgj):
    if Pi<Pj:
        Xi,Xj = PlanetPropertiestoLinearModelAmplitudes(T0i,Pi,massi,ei,pmgi,T0j,Pj,massj,ej,pmgj)
    else:
        Xj,Xi = PlanetPropertiestoLinearModelAmplitudes(T0j,Pj,massj,ej,pmgj,T0i,Pi,massi,ei,pmgi)
    return Xi[2:],Xj[2:]
###################################
def MultiplanetSystemLinearModelAmplitudes(Nplanets,Periods,T0s,masses,eccs,pomegas,**kwargs):
    """
    Compute amplitudes for linear model basis functions for user-specified planet parameters.

    Parameters
    ----------

    Nplanets    :   int
        The number of planets in the system.

    Periods     :   ndarray
        Periods of the planets.

    T0s         :   ndarray
        Times of first transit

    masses      :   ndarray
        Planet masses, in units of host-star mass.

    eccs        :   ndarray
        Eccentricities.

    pomegas     :   ndarray
        Longitudes of periapse.

    Returns
    -------

    :obj:`list` of ndarrays
        List of model amplitudes 
    """
    InteractionMatrix=kwargs.get("InteractionMatrix",np.ones((Nplanets,Nplanets),dtype=bool))
    for i in range(Nplanets):
        # No self-interactions!
        InteractionMatrix[i,i]=False

    Xs=[]
    for i in range(Nplanets):
        lenXi = 2 + 3 * np.sum(InteractionMatrix[i])
        Xi = np.zeros(lenXi)
        Xi[0] = T0s[i]
        Xi[1] = Periods[i]
        Xs.append(Xi)
    
    for i in range(Nplanets):
        Pi = Periods[i]
        T0i = T0s[i]
        massi=masses[i]
        ei=eccs[i]
        pmgi=pomegas[i]
        for j in range(i+1,Nplanets):
            if InteractionMatrix[i,j] or InteractionMatrix[j,i]:
                Pj = Periods[j]
                T0j = T0s[j]
                massj = masses[j]
                ej = eccs[j]
                pmgj=pomegas[j]

                Xi,Xj = get_ttv_model_amplitudes(Pi,Pj,T0i,T0j,massi,massj,ei,ej,pmgi,pmgj)

                if InteractionMatrix[i,j]:
                    Jindex=2+3*np.sum(InteractionMatrix[i,:j])
                    Xs[i][Jindex:Jindex+3]=Xi
                if InteractionMatrix[j,i]:
                    Iindex=2+3*np.sum(InteractionMatrix[j,:i])
                    Xs[j][Iindex:Iindex+3]=Xj
            else:
                pass
    return Xs
###################################
class PlanetTransitObservations(object):
    """
    Object to store transit timing measurement information.

    """
    def __init__(self,transit_numbers,times,uncertainties):
        """
        Parameters
        ----------

        transit_numbers : ndarray
            List that numbers the series of transit observations

        times           : ndarray
            List of transit mid-times

        uncertainties   : ndarray
            List of tramsit mid-time measurement :math:`1\sigma` uncertainties

        """
        self._transit_numbers=transit_numbers
        self._times=times
        self._uncertainties = uncertainties
        self._Ntransits = len(times)
        self._mask = np.ones(self._Ntransits,dtype=bool)

    @classmethod
    def from_times_only(cls,times,unc=0):
        """
        Initialize transit observation object diectly from list of transit times.

        Return a :class:`PlanetTransitObservations` object initialized from a list of transit times.
        Transit numbers are automatically assigned sequentially. Uniform uncertainties are
        assigned to the observations according to the user-specified `unc`.
        
        Parameters
        ----------

        times : array_like
            List of transit mid-times
        unc : float, optional
            Uncertainty assigned to transit observations

        """
        Ntr = len(times)
        nums = np.arange(Ntr)
        sigma = unc * np.ones(Ntr)
        return cls(nums,times,sigma)

    def function_mask(self,func):
        """
        Mask out transit times by applying function `func` to the times.

        Parameters
        ----------

        func : callable
            Function to mask times. `func` should return `False` for times that are
            to be masked out. Otherwise `func` should return `True`.
        """
        self._mask = np.array(list(map(func,self._times)))

    def unmask(self):
        self._mask = np.ones(self._Ntransits,dtype=bool)

    @property
    def times(self):
        """
        ndarray: List of transit mid-times
        """
        return self._times[self._mask]    
    @property
    def transit_numbers(self):
        """
        ndarray: List of transit epoch numbers.
        """
        return self._transit_numbers[self._mask]    
    @property
    def uncertainties(self):
        """
        ndarray: List of :math:`1\sigma` transit mid-time measurement uncertainties.
        """
        return self._uncertainties[self._mask]    

    @uncertainties.setter
    def uncertainties(self,value):
        self._uncertainties[self._mask] = value   

    @property
    def Ntransits(self):
        return len(self.times)
    
    @property
    def weighted_obs_vector(self):
        return self.times / self.uncertainties
    
    def basis_function_matrix(self):
        """ 
        Generate the basis function matrix for a linear transit ephemeris.

        Returns
        -------
        ndarray
            basis function matrix
        """
        constant_basis = np.ones(self.Ntransits)
        return np.vstack(( constant_basis , self.transit_numbers)).T
    def linear_fit_design_matrix(self):
        """ 
        Generate the design matrix for a linear transit ephemeris.
        """
        constant_basis = np.ones(self.Ntransits)
        sigma = self.uncertainties
        design_matrix = np.vstack(( constant_basis / sigma , self.transit_numbers / sigma )).T
        return design_matrix
    
    def linear_best_fit(self):
        """ 
        Determine the best-fit period and initial transit time for a linear transit ephemeris.
        """
        sigma = self.uncertainties
        y = self.weighted_obs_vector
        A = self.linear_fit_design_matrix()
        fitresults = np.linalg.lstsq(A,y,rcond=-1)
        return fitresults[0]
    def linear_fit_residuals(self):
        """
        Return the transit timing residuals of a best-fit linear transit ephemeris.
        """
        M = self.basis_function_matrix()
        best = self.linear_best_fit()
        return self.times - M.dot(best)

class transit_times_model(OrderedDict):
    """
    Class representing a linear model for a single set of transit times.

    Attributes
    ----------
    observations : :obj:`PlanetTransitObservations` 
        Transit observations to model.
    parameter_bounds  : :obj:`OrderedDict`
        A dictionary that contains bounding intervals
        for each linear model amplitude 
    MaxTransitNumber : int
        Largest transit epoch number of the transits recorded in
        'observations'. This is used when generating new basis 
        functions to add to the model. 
    """
    def __init__(self,observations,suffix=''):
        super(transit_times_model,self).__init__()
        self.observations = observations
        self.parameter_bounds = OrderedDict()
        self.MaxTransitNumber = observations._transit_numbers.max()
        self['T0{}'.format(suffix)] = np.ones(observations._Ntransits)
        self['P{}'.format(suffix)] = observations._transit_numbers
        self.parameter_bounds['P{}'.format(suffix)] = (0,np.inf)
        self._suffix = suffix
    def __reduce__(self):
        """Helps the class play nice with pickle_. 

        .. _pickle: https://docs.python.org/3/library/pickle.html#object.__reduce__"""
        red = (self.__class__, (self.observations,
                                self._suffix),
                                None,None,iter(self.items()))
        return red

   
    def __getitem__(self,key):
        return super().__getitem__(key)[self.mask]
       
    def __setitem__(self,key,val):
        if key not in self.parameter_bounds.keys():  
            self.parameter_bounds[key]=(-np.inf,np.inf)
        super().__setitem__(key,val)
        
    @property
    def mask(self):
        return self.observations._mask

    @property
    def Nrow(self):
        """int: Number of basis function matrix rows"""
        return len(self)

    @property
    def Ncol(self):
        """int: Number of basis function matrix columns"""
        return self.observations.Ntransits

    @property
    def basis_function_matrix(self):
        """Return the basis function matrix."""
        return np.vstack([val[self.mask] for val in self.values()]).T
    
    def design_matrix(self):
        """Return the design matrix."""
        sigma_vec = self.observations.uncertainties
        Atranspose = np.transpose(self.basis_function_matrix) / sigma_vec
        return np.transpose(Atranspose)

    def cov_inv(self):
        """Return the inverese covariance matrix"""
        A = self.design_matrix()
        return np.transpose(A).dot(A)

    def cov(self):
        """Return the covariance matrix."""
        return np.linalg.inv(self.cov_inv())

    def list_columns(self):
        """List the column labels of the basis function matrix"""
        return [key for key in self.keys()]
    
    def weighted_obs_vector(self):
        """Observation times weighed by uncertainties.
           
           Used primarily in least-squares fitting along with design
           matrix. 
        """
        return self.observations.times / self.observations.uncertainties

    def function_mask(self,maskfn):
        """Mask the underlying observations using function 'maskfn'
        Arguments
        ---------
        maskfn : callable
            Function of observationt time used to mask observations. 
            Observations times for which 'maskfn' returns 'False' are 
            masked out, otherwise they are included.
        """
        if not callable(maskfn):
            raise TypeError("'maskfn' must be callable.")
        self.observations.function_mask(maskfn)

    def best_fit(self,full_output=False):
        """Compute the best-fit transit model amplitudes
        subject to the constraints set in 'parameter_bounds'
        attribute.
        
        Arguments
        ---------
        full_output : bool (optional)
            If 'True', return the 'OptimizeResult' object generated
            by scipy.lsq_linear along with the default dictionary
            containing best-fit amplitudes

        Returns
        -------
        dictionary :
            A dictionary containing the best-fit model amplitudes.
        """
        A = self.design_matrix()
        y = self.weighted_obs_vector()
        lb,ub = np.transpose([bounds for bounds in self.parameter_bounds.values()])
        min_result = lsq_linear(A,y,bounds = (lb,ub))
        best_dict = {key:val for key,val in  zip(self.keys(),min_result.x)}
        if full_output:
            return best_dict,min_result
        return best_dict

    def best_fit_vec(self):
        """Get best-fit transit model amplitudes as a vector.

        Returns
        -------
        ndarray : 
            Vector representing the best-fit TTV model amplitudes
        """
        bfdict = self.best_fit()
        return np.array([bfdict[x] for x in self.list_columns()])

    def residuals(self):
        """Return the normalized residuals of the best-fit solution"""
        best_dict,min_result = self.best_fit(full_output=True)
        return min_result.fun
    def chi_squared(self,per_dof=False):
        """Return the chi-squared value of the best-fit solution.

        Arguments
        ---------
        per_dof : bool (optional)
            If true, return the chi-sqaured divided by the 
            number of degrees of freedom (i.e., the number
            of observations minus the number of model 
            parameters). The default value is False.
        """
        resids = self.residuals()
        chisq = resids.dot(resids)
        if per_dof:
            dof = self.Ncol - self.Nrow
            return chisq / dof
        return chisq

    def Delta_BIC(self):
        """
        Return the difference in Bayesian information criteria (BICs)
        between a purely linear transit time ephemeris and the full 
        transit time models.
        """
        chisq = self.chi_squared()
        penalty_term=np.log(self.Ncol)*self.Nrow
        BIC_full = chisq + penalty_term

        line_fit_resids = self.observations.linear_fit_residuals()
        chisq_linear_fit = line_fit_resids.dot(line_fit_resids)
        linear_fit_penalty = np.log(self.Ncol) * 2
        BIC_linear = chisq_linear_fit + linear_fit_penalty
        
        return BIC_linear - BIC_full
        

class TransitTimesLinearModels(object):
    """
    Object representing a collection of transit time linear models in a system of interacting planets.

   Parameters
   ----------
   observations_list : :obj:`list` of :obj:`PlanetTransitObservations`
       Set of transit observations to model.

    Keyword Arguments
    -----------------
    periods : :obj:`list` of floats
        Orbital periods of the transiting planets. Periods are determined
        by an initial least-squares fit if they are not supplied as a 
        keyword argument.
    T0s : :obj:`list` of floats
        Inital times of transits. Determined by a least-squares fit
        when not supplied as a keyword argument
    max_period_ratio : float
        Maximum period ratio for which planet-planet interactions are 
        included in the analytic TTV models. Default value is infinite.
    planet_names : str or list of str
        Names to label each planet with. The planet names appear as
        suffixes on basis function labels. 

    Attributes
    ----------
    observations : list
        A list of transit time observation objects representing the transit observations for a system
    
    basis_function_matrices: list
        List of each planet's transit itme basis function matrix
    
    periods : ndarray
        Best-fit periods of all planets

    T0s : ndarray
        Best-fit initial times of transit for all planets

    best_fit : :obj:`list` of ndarray
        List of ndarray best fit amplitudes of each planets' TTV basis functions.

    covariance_matrices : :obj:`list` of ndarray
        List of ndarray covariance matrices for each planets' TTV basis functions.

    """

    def __init__(self,observations_list,**kwargs):
        self.observations=observations_list
        for obs in self.observations:
            errmsg1 = "'TransitObservations' contains transits with negative transit numbers. Please re-number transits." 
            assert np.all(obs.transit_numbers>=0), errmsg1

        planet_names = kwargs.get("planet_names",["{}".format(i) for i in range(len(observations_list))])
        assert len(planet_names) == len(self.observations),\
            "Planet name string '{}' lengh does not match number of observations ({:d})".format(planet_names,len(self.observations))
        self.planet_names = planet_names
        self.planet_name_dict = {planet_name:i for i,planet_name in enumerate(planet_names)}

        periods=kwargs.get("periods",None)
        T0s =kwargs.get("T0s",None)
        max_period_ratio=kwargs.get("max_period_ratio",np.inf)
        if periods is None or T0s is None:
            initial_linear_fit_data = np.array([obs.linear_best_fit() for obs in self.observations ])
            if periods is None:
                periods = initial_linear_fit_data[:,1] 
            if T0s is None:
                T0s = initial_linear_fit_data[:,0]
        self.T0s = T0s
        self.periods = periods
        self.models = [ transit_times_model(obs,suffix = self.planet_names[i]) for i,obs in enumerate(self.observations) ]
        self._maximum_interaction_period_ratio = max_period_ratio
        self._interaction_matrix = SetupInteractionMatrixWithMaxPeriodRatio(self.periods,self.maximum_interaction_period_ratio)
        self.generate_basis_functions()


    def basis_function_matrices(self):
        """list : List containing the basis function matrix of each planet"""
        return [model.basis_function_matrix for model in self.models]

    def reset(self):
        """
        Reset TTV model.

        All TTV basis functions are erased and a linear ephemeris is re-fit to each planet's
        transit times. The interaction matrix is reset so that all pair-wise interactions
        are considered. 
        """
        initial_linear_fit_data = np.array([obs.linear_best_fit() for obs in self.observations ])
        self.T0s = initial_linear_fit_data[:,0]
        self.periods = initial_linear_fit_data[:,1]
        self.models = [ transit_times_model(obs) for obs in self.observations ]
        self._maximum_interaction_period_ratio = np.infty
        self._interaction_matrix = SetupInteractionMatrixWithMaxPeriodRatio(self.periods,self.maximum_interaction_period_ratio)

    @property 
    def interaction_matrix(self):
        """
        ndarray: Matrix with elements that record whether pair-wise interactions are 
        in each planet's set of TTV basis functions. If :math:`I_{ij}=` True, then 
        the basis functions accounting for perturbations by planet j on
        planet *i* are included in the model for planet i's TTV.
        """
        return self._interaction_matrix

    @interaction_matrix.setter
    def interaction_matrix(self,value):
        self._interaction_matrix = value

    @property
    def maximum_interaction_period_ratio(self):
        """
        float: Maximum period ratio above which planet-planet interactions are ignored in the TTV model.
        """
        return self._maximum_interaction_period_ratio

    @maximum_interaction_period_ratio.setter
    def maximum_interaction_period_ratio(self,value):
       self.interaction_matrix = SetupInteractionMatrixWithMaxPeriodRatio(self.periods,value)
       self._maximum_interaction_period_ratio =  value
    
    @property
    def N(self):
        """int: Number of planets with transit observations."""
        return len(self.observations)


    def design_matrices(self):
        """:obj:`list` of ndarrays: Design matrices for each planet's TTV model."""
        
        return [model.design_matrix() for model in self.models]


    def covariance_matrices(self):
        """:obj:`list` of ndarrays: Covariance matrices of each planet's TTV model."""
        return [model.cov() for model in self.models]
    
    def best_fits(self):
        """:obj:`list` of ndarrays: best-fit model amplitudes for each planet's TTV model."""
        best = dict()
        for model in self.models:
            best.update(model.best_fit())
        return best
        
    def chi_squareds(self,per_dof=False):
        """Return the chi-squareds of each transit time model.

        Arguments
        ---------
        per_dof : bool (optional)
            If true, return the chi-sqaured divided by the 
            number of degrees of freedom (i.e., the number
            of observations minus the number of model 
            parameters). The default value is False.

        Returns
        -------
        ndarray : 
            Chi-squared value of each planet's transit time
            model.
        """
        return np.array([mdl.chi_squared(per_dof) for mdl in self.models])

    def Delta_BICs(self):
        """Return the Delta-BIC of each transit model relative
        to linear ephemerides"""
        return np.array([mdl.Delta_BIC() for mdl in self.models])

    def generate_basis_functions(self,second_order_resonances=None):
        """Generate TTV basis functions and update planets' TTV models."""
        for name_i,i in self.planet_name_dict.items():
            per_i = self.periods[i]
            T0_i = self.T0s[i]
            model  = self.models[i]
            Ntrans = model.MaxTransitNumber+1
            for name_j,j in self.planet_name_dict.items():
                if self.interaction_matrix[i,j]:
                    per_j = self.periods[j]
                    T0_j = self.T0s[j]
                    bf_mtrx = get_ttv_basis_function_matrix(
                            per_i, per_j, T0_i, T0_j,Ntrans
                            )
                    bf_mtrx = bf_mtrx[model.observations._transit_numbers]
                    model['dt0_{}{}'.format(name_i,name_j)] = bf_mtrx[:,0]
                    model.parameter_bounds['dt0_{}{}'.format(name_i,name_j)] = (0,1)
                    model['dt1x_{}{}'.format(name_i,name_j)] = bf_mtrx[:,1]
                    model['dt1y_{}{}'.format(name_i,name_j)] = bf_mtrx[:,2]
                           
    def add_second_order_resonance(self,planet1,planet2):
        """Add basis-functions for a second-order resonance.

        Arguments
        ---------
        planet1 : str or int
            Name or index of the first planet
        planet2 : str or int
            Name or index of the second planet
        """

        if type(planet1) is str:
            i1str = planet1
            i1 = self.planet_name_dict[planet1]
        elif type(planet1) is int:
            i1str = list(self.planet_name_dict.keys())[planet1]
            i1 = planet1
        else:
            raise ValueError("'planet1' must be of type 'int' or 'str'")

        if type(planet2) is str:
            i2str = planet2
            i2 = self.planet_name_dict[planet2]
        elif type(planet2) is int:
            i2str = list(self.planet_name_dict.keys())[planet2]
            i2 = planet2
        else:
            raise ValueError("'planet2' must be of type 'int' or 'str'")
        
        p1 = self.periods[i1]
        p2 = self.periods[i2]
        T01 = self.T0s[i1]
        T02 = self.T0s[i2]
        Ntr1 = self.models[i1].MaxTransitNumber + 1
        Ntr2 = self.models[i2].MaxTransitNumber + 1
        if p1 < p2:
            dt2_1 = dt2_InnerPlanet(p1,p2,T01,T02,Ntr1).astype(float)
            dt2_2 = dt2_OuterPlanet(p1,p2,T01,T02,Ntr2).astype(float)
        else:
            dt2_1 = dt2_OuterPlanet(p2,p1,T02,T01,Ntr1).astype(float)
            dt2_2 = dt2_InnerPlanet(p2,p1,T02,T01,Ntr2).astype(float)

        self.models[i1]['dt2x_{}{}'.format(i1str,i2str)] = dt2_1[:,0]
        self.models[i1]['dt2y_{}{}'.format(i1str,i2str)] = dt2_1[:,1]
        self.models[i2]['dt2x_{}{}'.format(i2str,i1str)] = dt2_2[:,0]
        self.models[i2]['dt2y_{}{}'.format(i2str,i1str)] = dt2_2[:,1]

    def update_fits(self):
        """
        Compute the best-fit periods using current model then re-compute
        model basis functions.
        """
        fit_dict = self.best_fits()
        self.periods = [fit_dict['P{}'.format(name)] for name in self.planet_names]
        self.generate_basis_functions()

    def compute_ttv_significance(self):
        Sigma = self.covariance_matrices
        mu = self.best_fits
        significance_in_sigmas=[]
        for i in range(self.N):
            if len(mu[i] >2):
                muTTV = mu[i][2:]
                SigmaTTVinv= np.linalg.inv(Sigma[i][2:,2:])
                chisquared = muTTV.dot(SigmaTTVinv.dot(muTTV))
                dof = len(muTTV)
                sigma=chiSquared_to_sigmas(chisquared,dof)
                significance_in_sigmas.append(sigma)
            else:
                significance_in_sigmas.append(0)
        return significance_in_sigmas

    def mask_observations(self,maskfn):
        """Mask transit observations using function 'maskfn'
        Arguments
        ---------
        maskfn : callable
            Function of observation time used to mask observations. 
            Observations times for which 'maskfn' returns 'False' are 
            masked out, otherwise they are included.
        """
        if not callable(maskfn):
            raise TypeError("'maskfn' must be callable.")
        for model in self.models:
            model.function_mask(maskfn)

def interactionIndicies(LMsystem,i,j):
    i_matrix = LMsystem.interaction_matrix
    if  i_matrix[i,j]:
        k0 = 2 + 3 * np.sum( i_matrix[i,:j])
        i_indices = k0 + np.arange(3,dtype=int)
    else:
        i_indices = []

    if i_matrix[j,i]:
        l0 = 2 + 3 * np.sum( i_matrix[j,:i])
        j_indices = l0 + np.arange(3,dtype=int)
    else:
        j_indices = []
    
    return i_indices, j_indices

def chiSquared_to_sigmas(chi2,dof):
    """
    Convert a chi-squared value to a confidence level, in terms of 'sigmas'

    Arguments
    ---------
    
    chi2 : float
        Chi-squared value

    dof  : int
        Number of degrees of freedom

    Returns
    -------
    
    float
        The 'sigma's with the same confidence level for a 1D Gaussian
    """
    p = gammainc(0.5 * dof, 0.5 * chi2)
    return np.sqrt( 2 * gammaincinv( 0.5 , p ) )
#############################################



"""Module for computing upper limits on the mass of any unseen companions to an observed singly-transiting planet."""
import numpy as np
from scipy.optimize import brenth,nnls
from scipy.special import erf
from scipy.optimize import minimize,LinearConstraint
from scipy.integrate import trapezoid as trapz
from .ttv_basis_functions import dt0_InnerPlanet,dt0_OuterPlanet
from .ttv2fast2furious import PlanetTransitObservations

def PerturberPeriodPhaseToBestSigmaChiSquared(Ppert,phi,TransitObservations, PlanetData = None,full_output=False):
    """
    Convert a hypothetical perturber period and phase to a mean and std. deviation
    of a gaussian distribution describing the perturber mass conditional posterior.

    Parameters
    ----------
    Ppert : real
        Period of the pertuber
    phi : real
        Orbital phase of perturber
    TransitData : PlanetTransitObservations
        Observation data on transiting planet.
    PlanetData : list of reals (optional)
        List containing transiting planet's T0 and Period.
    full_output : bool (optional)
        Include a dictionary in the function output that contains 
         the full best fit, covariance matrix, design matrix, 
         and weighted observations vector.

    Returns
    -------
    results : tuple 
        Returns the best-fit mass, mass distribution 'sigma' and the chi-squared of
        the timing residuals.
    """

    transit_num = TransitObservations.transit_numbers
    transit_time = TransitObservations.times
    transit_unc = TransitObservations.uncertainties

    assert np.all(transit_num>=0), " 'TransitObservations' contains transits with negative transit numbers. Please re-number transits." 
    assert np.all(transit_time>0), " 'TransitObservations' contains transits with negative transit time. Negative transit times are not supported since a non-negative least squares algorithm is used to determine the best-fit basis function amplitudes." 

    yvec = transit_time / transit_unc
    Ntransits = int(np.max(transit_num))
    Tpert = Ppert * phi / (2*np.pi)


    if PlanetData is None:
        Ndata = len(transit_num)
        T0,P = TransitObservations.linear_best_fit()
    else:
        T0,P = PlanetData

    if Ppert > P:
        dt0_basis_fn = dt0_InnerPlanet(P,Ppert,T0,Tpert,Ntransits+1)
    else:
        dt0_basis_fn = dt0_OuterPlanet(P,Ppert,T0,Tpert,Ntransits+1)

    bfMatrix = np.vstack([np.ones(Ntransits+1) , np.arange(Ntransits+1) , dt0_basis_fn]).T

    design_Matrix = np.array([bfMatrix[i]/transit_unc[ni]  for ni,i in enumerate(np.array(transit_num,dtype=int))])

    Sigma_matrix = np.linalg.inv(design_Matrix.T.dot(design_Matrix))
    # New method for obtaining best fit using scipy's 'nnls' algorithm
    # Use of this algorithm requires all transit times to be non-negative 
    best,chisq = nnls(design_Matrix,yvec)

    if full_output:
        return best[2], np.sqrt(Sigma_matrix[2,2]) ,chisq ,dict({'BestFit':best,'CovarianceMatrix':Sigma_matrix,'DesignMatrix':design_Matrix,'WeightedObsVec':yvec,'BasisFunctionMatrix':bfMatrix})

    return best[2], np.sqrt(Sigma_matrix[2,2]) ,chisq

def q_integrand(mup,mbest,sigma):
    """
    Compute the value of the integrand in the integral over phase that determines q(m).

    Arguments
    ---------

    mup : real
        The mass upper limit for which q is being computed
    mbest : real
        The best-fit mass (mhat) which sets the mean of the gaussian mass distribution. Depends on phase phi.
    sigma : real
        The std. deviation parameter of the gaussian mass distribution. Depends on phase phi.

    Returns
    -------

    real :
        The value of the integrand
    """
    num = erf(mbest / sigma / np.sqrt(2)) +  erf((mup-mbest) / sigma / np.sqrt(2))
    denom = 1 + erf(mbest / sigma / np.sqrt(2))
    return num / denom

def UnseenPerturberMassUpperLimit(Ppert,confidence_levels,TransitObservations ,Nphi = 50,Mmax0 = 3e-3,PlanetData = None):
    """
    Compute mass upper limit(s) on a potential perturber at a given orbital period using transit data.
    Marginalizes over possible orbital phases of the perturber.

    Parameters
    ----------
    Ppert : real
        The orbital period of the hypothetical perturbing planet
    confidence_levels : array-like
        The confidence level(s) for which to return mass upper limits.
    TransitObservations : PlanetTransitObservations
        Observation data on transiting planet
    Nphi : int (optional)
        Number of perturber phase points used to compute integrals for marginalizing over phase.
        Default value is Nphi=50.
    Mmax0 : real (optional)
        Initial guess of mass upper limit for root-finding purposes. All mass upper-limits are
        guessed to fall between 0 and Mmax0.

    PlanetData : list (optional)
        List containing [T0,P] where T0 is the initial transit time and P is period 
         of the traniting planet. If these values are not supplied they are computed
         from the transit data.
    Returns
    -------
    limits : array-like
        List of mass upper limits and the given confidence levels.
    """



    phases = np.linspace(-np.pi,np.pi,Nphi)
    mbest,sigma,chisq = np.array([PerturberPeriodPhaseToBestSigmaChiSquared(Ppert,phase,TransitObservations,PlanetData=PlanetData) for phase in phases]).T
    dchisq = chisq - np.mean(chisq)
    q_of_mup = lambda mup: trapz( q_integrand(mup,mbest,sigma) * np.exp(-0.5 * dchisq),phases) / trapz(np.exp(-0.5 * dchisq),phases)

    mups = []

    for cl in np.atleast_1d(confidence_levels):
        assert (cl < 1), "%.2f is not a valid confidence level between 0 and 1!"%cl
        while q_of_mup(Mmax0) - cl < 0:
            Mmax0 *=2
        mups.append(brenth(lambda x: q_of_mup(x) - cl,0,Mmax0))

    return np.array(mups)



import pandas as pd
import os
from . import PlanetTransitObservations

_pymodulepath = os.path.dirname(__file__)
_RoweTableFile = _pymodulepath+"/../data/RoweTableDataframe.pkl"
_fileExists = os.path.isfile(_RoweTableFile)
if not _fileExists:
    import urllib.request
    print("Downloading transit time table file...\n\t (This may take a minute but you'll only need to do it once.)")
    _url = "https://www.cfa.harvard.edu/~shadden/RoweTableDataframe.pkl"
    os.mkdir(_pymodulepath+"/../data")
    try:
        urllib.request.urlretrieve(_url,_RoweTableFile)
        _fileExists = True
    except: 
        print("Table download failed.")
        print("ttv2fast2furious.kepler requires the data file: \n\t %s"%_RoweTableFile)
        print("The data file can be retrieved manually from %s"%_url)

if _fileExists:
    KOITransitTimeDataframe = pd.read_pickle(_RoweTableFile)
else:
    print("ttv2fast2furious.kepler requires the data file: \n\t %s"%_RoweTableFile)
    print("The data file can be retrieved manually from %s"%_url)
def KOISystemObservations(KOINumber):
    """
    Retrieve the transit time observations for a KOI system

    Parameters
    ----------
    KOINumber : int or str
        The KOI number of the system to retrieve

    Returns
    -------
    observations : dict of PlanetTransitObservations
        Transit time observations of the KOIs in the system.
    """
    if not _fileExists:
        print("Table data file required.")
        return 
    KOI = int(KOINumber)
    if KOI not in KOITransitTimeDataframe.KOI:
        print("No entry found for KOI%d"%KOI)
        return None
    observations = dict()
    koi_system_dframe = KOITransitTimeDataframe.query("`KOI_System`==@KOI")
    for koi,obs in koi_system_dframe.groupby('KOI'):
        observations.update({
            koi:PlanetTransitObservations(obs.TransitNumber.values,obs.TransitTime.values,obs.eTTV.values)
            })
    return observations



# -*- coding: utf-8 -*-
"""Tools for TTV analysis and parameter inference."""
# Find suffix
import sysconfig
_suffix = sysconfig.get_config_var('EXT_SUFFIX')
if _suffix is None:
    _suffix = ".so"

# Import shared library
import os
import warnings
pymodulepath = os.path.dirname(__file__)
from ctypes import cdll
_clibttv2fast2furious = cdll.LoadLibrary(pymodulepath+"/../libttv2fast2furious"+_suffix)

from . import companion_limits 
from .ttv2fast2furious import PlanetTransitObservations,TransitTimesLinearModels
from .ttv2fast2furious import MultiplanetSystemBasisFunctionMatrices
from .ttv2fast2furious import MultiplanetSystemLinearModelAmplitudes

__all__ = ['MultiplanetSystemLinearModelAmplitudes','MultiplanetSystemBasisFunctionMatrices','PlanetTransitObservations','TransitTimesLinearModels','companion_limits']



import subprocess
from setuptools import setup, Extension
from codecs import open
import os
import sys

import sysconfig
suffix = sysconfig.get_config_var('EXT_SUFFIX')
if suffix is None:
    suffix = ".so"
if sys.platform == 'darwin':
    from distutils import sysconfig
    vars = sysconfig.get_config_vars()
    vars['LDSHARED'] = vars['LDSHARED'].replace('-bundle', '-shared')

def get_gsl_config(flag):
    try:
        # Run pkg-config to retrieve the appropriate flags
        result = subprocess.check_output(['pkg-config', f'--{flag}', 'gsl'])
        return result.decode('utf-8').strip().split()
    except subprocess.CalledProcessError:
        print(f"Error: Unable to find GSL with pkg-config for flag {flag}")
        return []
    except FileNotFoundError:
        print("Error: pkg-config is not installed. Please install it.")
        return []

# Retrieve include and library paths dynamically
include_dirs = [dir.lstrip('-I') for dir in get_gsl_config('cflags')]
library_dirs = [dir.lstrip('-L') for dir in get_gsl_config('libs')]
libraries = ['gsl', 'gslcblas', 'm']  # GSL and additional required libraries

libttv2fast2furiousmodule = Extension(
    'libttv2fast2furious',
    sources=['src/basis_functions.c'],
    include_dirs=['src'] + include_dirs,  # Add 'src' and GSL include dirs
    library_dirs=library_dirs,
    libraries=libraries,
    define_macros=[('LIBTTV2FAST2FURIOUS', None)],
    extra_compile_args=[
        '-fstrict-aliasing',
        '-O3',
        '-std=c99',
        '-Wno-unknown-pragmas',
        '-DLIBTTV2FAST2FURIOUS',
        '-fPIC',
    ],
    extra_link_args=[],  # You can add extra linker arguments here if needed
)

setup(name='ttv2fast2furious',
    version='1.1.0',
    description='Open source tools for TTV analysis and parameter inference.',
    url='http://github.com/shadden/TTV2Fast2Furious',
    author='Sam Hadden',
    author_email='shadden1107@gmail.com',
    license='GPL',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Scientific/Engineering :: Astronomy',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
    keywords='astronomy astrophysics',
    packages=['ttv2fast2furious'],
    install_requires=['numpy', 'scipy','sympy','pandas'],
    include_package_data=True,
    ext_modules = [libttv2fast2furiousmodule],
    zip_safe=False)



"""Routines for computing the basis functions used in the TTV linear model."""

import numpy as np
from scipy.special import gammainc,gammaincinv
from scipy.special import ellipk,ellipkinc,ellipe,ellipeinc
import scipy.integrate as integrate
import os
from . import _clibttv2fast2furious
from ctypes import c_double

##################################
########## dl terms #############
##################################
# Equations A21 and A22 
def ellipkOsc(x,m):
    """
    Compute the oscillating part of elliptic function ?th kind,  K

    Args:
        x (real): elliptic function arugment
        m (real): elliptic function modulus

    Returns:
        real
    """
    return ellipkinc(x , m) - 2*x*ellipk(m)/(np.pi)
def ellipeOsc(x,m):
    """
    Compute the oscillating part of elliptic function ?th kind, E

    Args:
        x (real): elliptic function arugment
        m (real): elliptic function modulus

    Returns:
        real
    """
    return ellipeinc(x , m) - 2*x*ellipe(m)/(np.pi)

# oscillating part of integral (1+a^2-2a*cos(psi))^(-1/2)
def DinvOsc(alpha,psi):
    denom = 1-alpha
    m = -4*alpha / denom / denom
    psi=np.mod(psi,2*np.pi)
    return 2 * ellipkOsc(0.5 * psi , m) / denom

# oscillating part of integral d/da[(1+a^2-2a*cos(psi))^(-1/2)]
def dDdalphaInvOsc(alpha,psi):
    m = -4*alpha / (1-alpha) / (1-alpha)
    X = (1-alpha) * ellipeOsc(0.5 * psi,m)
    X = X - (1+alpha)* ellipkOsc(0.5 * psi,m)
    X = X + 2* alpha * np.sin(psi) / np.sqrt(1+alpha*alpha-2*alpha*np.cos(psi))
    X = X / alpha / (1-alpha) / (1+alpha)
    return X

# Equation A19 and A20
def A(alpha,psi):
    indirect= -3 * np.sin(psi) / alpha
    direct = 3 * DinvOsc(alpha,psi)  / alpha / alpha
    return direct + indirect
def B(alpha,psi):
    indirect = -2 * np.sqrt(alpha) * np.sin(psi)
    direct = 2 * np.sqrt(alpha) * dDdalphaInvOsc(alpha,psi)
    return  direct + indirect

# Equation A29
def A1(alpha,psi):
    indirect = 3 * np.sin(psi) / alpha / alpha
    direct = -3 * DinvOsc(alpha,psi) 
    return direct + indirect

def B1(alpha,psi):
    direct = -2 * (DinvOsc(alpha,psi) + alpha*dDdalphaInvOsc(alpha,psi))
    indirect = -2 * np.sin(psi) / alpha / alpha
    return direct + indirect


# inner planet lambda
def dlFn(alpha,psi):
    dnInv = 1.0/(alpha**(-1.5)-1.0)
    return dnInv*dnInv*A(alpha,psi) + dnInv*B(alpha,psi)
# outer planet lambda
def dl1Fn(alpha,psi):
    dnInv = 1.0/(alpha**(-1.5)-1.0)
    return dnInv*dnInv*A1(alpha,psi) + dnInv * B1(alpha,psi)


##################################
########## dz terms #############
##################################
_clibttv2fast2furious.dz1Re.restype = c_double
_clibttv2fast2furious.dz1Im.restype = c_double
_clibttv2fast2furious.dzRe.restype = c_double
_clibttv2fast2furious.dzIm.restype = c_double
def dzRe(alpha,t,t0=0,psi0=0,l0=0):
    return _clibttv2fast2furious.dzRe(c_double(alpha),c_double(t),c_double(t0),c_double(psi0),c_double(l0))
def dzIm(alpha,t,t0=0,psi0=0,l0=0):
    return _clibttv2fast2furious.dzIm(c_double(alpha),c_double(t),c_double(t0),c_double(psi0),c_double(l0))
def dz1Re(alpha,n1t,n1t0=0,psi0=0,l10=0):
    return _clibttv2fast2furious.dz1Re(c_double(alpha),c_double(n1t),c_double(n1t0),c_double(psi0),c_double(l10))
def dz1Im(alpha,n1t,n1t0=0,psi0=0,l10=0):
    return _clibttv2fast2furious.dz1Im(c_double(alpha),c_double(n1t),c_double(n1t0),c_double(psi0),c_double(l10))

########################################
########## ttv basis functions #########
########################################
def getTimesOfTransit(P,l0,N):
    return np.arange(1,N+1)*P - np.mod(l0,2*np.pi) * P / (2*np.pi)
def dt0_InnerPlanet(P,P1,T0,T10,Ntrans):
    """
    Compute the 0th order (in eccentricity) TTV basis function for an 
    inner planet with exterior perturber.

    Arguments
    ---------
        P : real
            The period of the planet
        P1 : real
            The period of the perturber
        T0 :  real
            Planet's time of initial transit
        T1 : real
            Perturbers time of initial transit
        Ntrans : int
            Number of transits to compute

    Returns
    -------

        ndarray :
            The basis function value at sequential transit times
    """

    # compute mean longitudes and synodic angle
    l = 2*np.pi * -T0 / P
    l1= 2*np.pi * -T10 / P1

    psi0=np.mod(l1-l ,2*np.pi)
    
    # get the unperturbed transit times
    TransitTimes0=getTimesOfTransit(P,l,Ntrans)
    TransitPhases0=2*np.pi * TransitTimes0 / P

    # get psi at the unperturbed transit times (i.e., when l=0)
    psi = np.mod(l1 + 2 * np.pi * TransitTimes0 / P1,2*np.pi)

    #get semi-major axis ratio
    alpha = np.power(P/P1,2./3.)
    
    # get list of delta-lambdas
    dl = dlFn(alpha,psi) 
    
    # get list of delta-z
    dzReArr = np.zeros(Ntrans)
    dzImArr = np.zeros(Ntrans)
    zlastRe=0.
    zlastIm=0.
    
    for i,t in enumerate(TransitPhases0):
        if i>0:
            zlastRe=zlastRe + dzRe(alpha,t,t0=TransitPhases0[i-1],psi0=psi0,l0=l)
            zlastIm=zlastIm + dzIm(alpha,t,t0=TransitPhases0[i-1],psi0=psi0,l0=l)
        else:
            zlastRe =  dzRe(alpha,t,psi0=psi0,l0=l)
            zlastIm =  dzIm(alpha,t,psi0=psi0,l0=l)
        dzReArr[i] = zlastRe
        dzImArr[i] = zlastIm
    dl = dl-np.mean(dl)
    dzImArr = dzImArr-np.mean(dzImArr)
    dzReArr = dzReArr-np.mean(dzReArr)
    dz = dzImArr
    ttv0 = -1 * P *  (dl - 2 * dz) / (2*np.pi)
    return ttv0
########################################
def dt0_OuterPlanet(P,P1,T0,T10,Ntrans):

    # compute mean longitudes and synodic angle
    l = 2*np.pi * -T0 / P
    l1= 2*np.pi * -T10 / P1

    psi0=np.mod(l1-l ,2*np.pi)
    
    # get the unperturbed transit times
    TransitTimes0=getTimesOfTransit(P1,l1,Ntrans)
    TransitPhases0=2*np.pi * TransitTimes0 / P1

    # get psi at the unperturbed transit times (i.e., when l1=0)
    psi = np.mod(l + 2 * np.pi * TransitTimes0 / P,2*np.pi)

    #get semi-major axis ratio
    alpha = np.power(P/P1,2./3.)
    
    # get list of delta-lambdas
    dl1A = dl1Fn(alpha,psi)

    
    # get list of delta-z
    dz1ReArr = np.zeros(Ntrans)
    dz1ImArr = np.zeros(Ntrans)
    z1lastRe=0.
    z1lastIm=0.
    
    for i,n1t in enumerate(TransitPhases0):
        if i>0:
            z1lastRe=z1lastRe + dz1Re(alpha,n1t,n1t0=TransitPhases0[i-1],psi0=psi0,l10=l1)
            z1lastIm=z1lastIm + dz1Im(alpha,n1t,n1t0=TransitPhases0[i-1],psi0=psi0,l10=l1)
        else:
            z1lastRe =  dz1Re(alpha,n1t,psi0=psi0,l10=l)
            z1lastIm =  dz1Im(alpha,n1t,psi0=psi0,l10=l)
        dz1ReArr[i] = z1lastRe
        dz1ImArr[i] = z1lastIm
    dl1A = dl1A-np.mean(dl1A)
    dz1ImArr = dz1ImArr-np.mean(dz1ImArr)
    dz1ReArr = dz1ReArr-np.mean(dz1ReArr)
    dz1 = dz1ImArr
    ttv = -1 * P1 *  (-dl1A - 2 * dz1) / (2*np.pi)
    return ttv
########################################

def get_nearest_firstorder(periodratio):
    return int(np.round((1-periodratio)**(-1)))
def get_superperiod(P,P1):
    j=get_nearest_firstorder(P/P1)
    SuperP = (j/P1 - (j-1)/P)**(-1)
    return SuperP

###########################################################
########## sympy for laplace coefficients #################
##########################################################
on_rtd = os.environ.get("READTHEDOCS", None) == "True"
if not on_rtd:
    from sympy import hyper,hyperexpand,S,binomial,diff,Subs,N
    _j = S("j")
    _alpha = S("alpha")
    _laplaceB= 2 * (-1)**_j *_alpha**_j * binomial(-1/S(2),_j)*hyper([1/S(2),_j+1/S(2)],[1+_j],_alpha*_alpha)
    _aDlaplaceB = _alpha * diff(_laplaceB,_alpha)
    _a2D2laplaceB = _alpha * _alpha * diff(_laplaceB,_alpha,_alpha)
    fCoeffInExprn = -_j * _laplaceB - (1/S(2)) * _aDlaplaceB
    fCoeffOutExprn = (_j + 1/S(2)) * _laplaceB + (1/S(2)) * _aDlaplaceB
    f49Coeff = -1 * (2 * _j * ( 1 + 2 *_j) * _laplaceB + (2 + 4 * _j) * _aDlaplaceB +  _a2D2laplaceB) / 4

def get_fCoeffs(j,alpha):
    fCoeffIn = N(Subs(fCoeffInExprn,[_j,_alpha],[j,alpha])) 
    fCoeffOut = N(Subs(fCoeffOutExprn,[_j,_alpha],[j-1,alpha])) 
    # Add indirect term
    if j==2:
     # Evaluated exactly on resonance, indirect contribution to 
     # fCoeffOut is the same for inner and outer planet. It is 
     # different for different values of aIn/aOut but I'm ignoring
     # this for now.
        fCoeffOut = fCoeffOut - 2.0**(1./3.)
    return float(fCoeffIn),float(fCoeffOut)

########################################
########### 2nd order res ##############
########################################
def get_gamma(k,alpha):
    f49 =  N(Subs(f49Coeff, [_j,_alpha] , [k-1,alpha]))
    j = int(np.ceil(k/2))
    f27,f31 = get_fCoeffs(j,alpha)
    # Eq. 58, Hadden & Lithwick '16
    gamma = f49 * (f27*f27 + f31*f31) / f27 / f31 / 2
    return gamma
def get_nearest_secondorder(Pin_by_Pout):
    return int(np.round(2 / (1 - Pin_by_Pout))) 
def get_second_order_superperiod(Pin,Pout):
    k = get_nearest_secondorder(Pin/Pout) 
    nIn = 2*np.pi / Pin
    nOut= 2*np.pi / Pout
    nRes = k*nOut - (k-2)*nIn
    return 2*np.pi / nRes
def dt2_InnerPlanet(P,P1,T0,T10,Ntrans):
    k=get_nearest_secondorder(P/P1)
    superP = get_second_order_superperiod(P,P1)
    superPhase = 2*np.pi * (k * (-T10/P1) - (k-2) * (-T0/P) )
    Times = T0 + np.arange(Ntrans) * P

    alpha = (P/P1)**(2./3.)
    alpha_2 = 1.0 / alpha / alpha
    gamma = get_gamma(k,alpha)
    n_K_2minusK = (1 / P1) / (k /P1 + (2-k) / P )

    # Eq. 60 of HL16
    S =  1.5 * (2-k) * alpha_2 * n_K_2minusK**2 * gamma * P / (np.pi)
    C = -1.5 * (2-k) * alpha_2 * n_K_2minusK**2 * gamma * P / (np.pi)
    superSin = S * np.sin(2*np.pi * Times / superP + superPhase)
    superCos = C * np.cos(2*np.pi * Times / superP + superPhase)
    basis_function_matrix=np.vstack((superSin,superCos)).T
    return basis_function_matrix

def dt2_OuterPlanet(P,P1,T0,T10,Ntrans):
    k=get_nearest_secondorder(P/P1)
    superP = get_second_order_superperiod(P,P1)
    superPhase = 2*np.pi * (k * (-T10/P1) - (k-2) * (-T0/P) )
    Times = T10 + np.arange(Ntrans) * P1
    alpha = (P/P1)**(2./3.)
    gamma = get_gamma(k,alpha)
    n_K_2minusK = (1 / P1) / (k /P1 + (2-k) / P )
    # Eq. 60 of HL16
    S =  1.5 * (k) * n_K_2minusK**2 * gamma * P1 / (np.pi)
    C = -1.5 * (k) * n_K_2minusK**2 * gamma * P1 / (np.pi)
    superSin = S * np.sin(2*np.pi * Times / superP + superPhase)
    superCos = C * np.cos(2*np.pi * Times / superP + superPhase)
    basis_function_matrix=np.vstack((superSin,superCos)).T
    return basis_function_matrix
########################################
